@@ -27,11 +27,17 @@
 #include <utime.h>
 #include "libbb.h"
 
-#define CONFIG_FEATURE_TAR_OLDGNU_COMPATABILITY
+#define CONFIG_FEATURE_TAR_OLDGNU_COMPATABILITY 1
+// #define CONFIG_FEATURE_TAR_GNU_EXTENSIONS
+
+#ifdef CONFIG_FEATURE_TAR_GNU_EXTENSIONS
+static char *longname = NULL;
+static char *linkname = NULL;
+#endif
 
 extern void seek_sub_file(FILE *src_stream, const int count);
 extern char *extract_archive(FILE *src_stream, FILE *out_stream, const file_header_t *file_entry,
- const int function, const char *prefix);
+                              const int function, const char *prefix);
 
 
 #ifdef L_archive_offset
@@ -75,8 +81,11 @@ char *extract_archive(FILE *src_stream, FILE *out_stream, const file_header_t *f
 {
 	FILE *dst_stream = NULL;
 	char *full_name = NULL;
+	char *full_link_name = NULL;
 	char *buffer = NULL;
 	struct utimbuf t;
+        int res;
+
 
 	/* prefix doesnt have to be a proper path it may prepend 
 	 * the filename as well */
@@ -93,9 +102,18 @@ char *extract_archive(FILE *src_stream, FILE *out_stream, const file_header_t *f
 		full_name = xmalloc(strlen(prefix) + strlen(path) + 1);
 		strcpy(full_name, prefix);
 		strcat(full_name, path);
+                if ( file_entry->link_name ){
+		   full_link_name = xmalloc(strlen(prefix) + strlen(file_entry->link_name) + 1);
+		   strcpy(full_link_name, prefix);
+		   strcat(full_link_name, file_entry->link_name);
+                }
 	} else {
 		full_name = strdup(file_entry->name);
+                if ( file_entry->link_name )
+		   full_link_name = strdup(file_entry->link_name);
 	}
+
+
 	if (function & extract_to_stdout) {
 		if (S_ISREG(file_entry->mode)) {
 			copy_file_chunk(src_stream, out_stream, file_entry->size);			
@@ -142,7 +160,7 @@ char *extract_archive(FILE *src_stream, FILE *out_stream, const file_header_t *f
 		switch(file_entry->mode & S_IFMT) {
 			case S_IFREG:
 				if (file_entry->link_name) { /* Found a cpio hard link */
-					if (link(file_entry->link_name, full_name) != 0) {
+					if (link(full_link_name, full_name) != 0) {
 						if ((function & extract_quiet) != extract_quiet) {
 							perror_msg("Cannot link from %s to '%s'",
 								file_entry->name, file_entry->link_name);
@@ -186,6 +204,9 @@ char *extract_archive(FILE *src_stream, FILE *out_stream, const file_header_t *f
 					return NULL;
 				}
 				break;
+                         default:
+                            perror_msg("Don't know how to handle %s", full_name);
+
 		}
 
 		/* Changing a symlink's properties normally changes the properties of the 
@@ -201,8 +222,8 @@ char *extract_archive(FILE *src_stream, FILE *out_stream, const file_header_t *f
 				t.modtime = file_entry->mtime;
 				utime(full_name, &t);
 			}
-			chmod(full_name, file_entry->mode);
 			chown(full_name, file_entry->uid, file_entry->gid);
+			chmod(full_name, file_entry->mode);
 		}
 	} else {
 		/* If we arent extracting data we have to skip it, 
@@ -224,6 +245,8 @@ char *extract_archive(FILE *src_stream, FILE *out_stream, const file_header_t *f
 	}
 
 	free(full_name);
+        if ( full_link_name )
+	    free(full_link_name);
 
 	return(NULL); /* Maybe we should say if failed */
 }
@@ -264,7 +287,6 @@ char *unarchive(FILE *src_stream, FILE *out_stream,
 			}
 
 		}
-		/* fprintf(stderr, __FUNCTION__ " got headers\n"); */
 
 		if (extract_flag == TRUE) {
 			/* fprintf(stderr, __FUNCTION__ " extract?\n"); */
@@ -398,6 +420,7 @@ file_header_t *get_header_cpio(FILE *src_stream)
  	static struct hardlinks *saved_hardlinks = NULL;
  	static int pending_hardlinks = 0;
 
+
 	if (pending_hardlinks) { /* Deal with any pending hardlinks */
 		struct hardlinks *tmp = saved_hardlinks, *oldtmp = NULL;
 		while (tmp) {
@@ -501,6 +524,7 @@ file_header_t *get_header_cpio(FILE *src_stream)
 #ifdef L_get_header_tar
 file_header_t *get_header_tar(FILE *tar_stream)
 {
+
 	union {
 		unsigned char raw[512];
 		struct {
@@ -566,9 +590,30 @@ file_header_t *get_header_tar(FILE *tar_stream)
 
 	/* convert to type'ed variables */
 	tar_entry = xcalloc(1, sizeof(file_header_t));
-	tar_entry->name = xstrdup(tar.formated.name);
 
+#ifdef CONFIG_FEATURE_TAR_GNU_EXTENSIONS
+        if (longname) {
+                tar_entry->name = longname;
+                longname = NULL;
+        }
+        else if (linkname) {
+                tar_entry->name = linkname;
+                linkname = NULL;
+        } else
+#endif
+        if (tar.formated.prefix[0] == 0) {
+                tar_entry->name = strdup(tar.formated.name);
+        } else {                                              
+                tar_entry->name = concat_path_file(tar.formated.prefix, tar.formated.name);
+        }
+
+	// tar_entry->name = xstrdup(tar.formated.name);
+
+/*
 	parse_mode(tar.formated.mode, &tar_entry->mode);
+*/
+        tar_entry->mode = 07777 & strtol(tar.formated.mode, NULL, 8);
+
 	tar_entry->uid   = strtol(tar.formated.uid, NULL, 8);
 	tar_entry->gid   = strtol(tar.formated.gid, NULL, 8);
 	tar_entry->size  = strtol(tar.formated.size, NULL, 8);
@@ -580,14 +625,19 @@ file_header_t *get_header_tar(FILE *tar_stream)
 
 	/* Fix mode, used by the old format */
 	switch (tar.formated.typeflag) {
+	/* hard links are detected as regular files with 0 size and a link name */
+	case '1':
+		tar_entry->mode |= S_IFREG ;
+		break;
+        case 0:
+        case '0':
+
 # ifdef CONFIG_FEATURE_TAR_OLDGNU_COMPATABILITY
-	case 0:
-	case '0':
 		if (last_char_is(tar_entry->name, '/')) {
 			tar_entry->mode |= S_IFDIR;
-		} else {
+		} else 
+# endif
 			tar_entry->mode |= S_IFREG;
-		}
 		break;
 	case '2':
 		tar_entry->mode |= S_IFLNK;
@@ -604,36 +654,35 @@ file_header_t *get_header_tar(FILE *tar_stream)
 	case '6':
 		tar_entry->mode |= S_IFIFO;
 		break;
-# endif
-	/* hard links are detected as regular files with 0 size and a link name */
-	case '1':
-		tar_entry->mode &= (S_IFREG | 07777);
-		break;
 # ifdef CONFIG_FEATURE_TAR_GNU_EXTENSIONS
 	case 'L': {
 			longname = xmalloc(tar_entry->size + 1);
-			archive_xread_all(archive_handle, longname, tar_entry->size);
+			archive_xread_all((int) tar_stream, longname, tar_entry->size);
 			longname[tar_entry->size] = '\0';
-			archive_handle->offset += tar_entry->size;
+			archive_offset += tar_entry->size;
 
-			return(get_header_tar(archive_handle));
+			return(get_header_tar(tar_stream));
 		}
 	case 'K': {
 			linkname = xmalloc(tar_entry->size + 1);
-			archive_xread_all(archive_handle, linkname, tar_entry->size);
+			archive_xread_all((int) tar_stream, longname, tar_entry->size);
 			linkname[tar_entry->size] = '\0';
-			archive_handle->offset += tar_entry->size;
+			archive_offset += tar_entry->size;
 
 			tar_entry->name = linkname;
-			return(get_header_tar(archive_handle));
+			return(get_header_tar(tar_stream));
 		}
 	case 'D':
 	case 'M':
 	case 'N':
 	case 'S':
 	case 'V':
-		bb_error_msg("Ignoring GNU extension type %c", tar.formated.typeflag);
+		perror_msg("Ignoring GNU extension type %c", tar.formated.typeflag);
 # endif
+        default:
+                perror_msg("Unknown typeflag: 0x%x", tar.formated.typeflag);
+                break;
+
 	}
 
 	return(tar_entry);
